

/*	Copyright (C) G. Jespersen
	See Licence.txt.*/

#include "Stdafx.h"
#include "Khan.h"
#include "History.h"
#include "math.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//////////////////////////////////////////////////////////////
// CHistoryItem

CHistoryItem::CHistoryItem()
{
	Zero();
}

CHistoryItem::~CHistoryItem()
{
}

CHistoryItem::CHistoryItem(const CHistoryItem & item)
{
	Copy(item);
}

void CHistoryItem::operator=(const CHistoryItem & item)
{
	Copy(item);
}

// Check if an exact duplicate.
BOOL CHistoryItem::operator==(const CHistoryItem & item)
{
	if(m_video			== item.m_video			&&
	   m_lastView		== item.m_lastView		&&
	   m_totalViewTime	== item.m_totalViewTime	&&
	   m_inMultiTopics	== item.m_inMultiTopics	)
		return TRUE;
	else
		return FALSE;
}

// Check if not an exact duplicate.
BOOL CHistoryItem::operator!=(const CHistoryItem & item)
{
	return !(*this == item);
}

// Copy all members.
void CHistoryItem::Copy(const CHistoryItem & item)
{
	m_video			= item.m_video;
	m_lastView		= item.m_lastView;
	m_totalViewTime = item.m_totalViewTime;
	m_inMultiTopics = item.m_inMultiTopics;
}

void CHistoryItem::Zero()
{
	m_video.Zero();
//	m_lastView = GetDefaultTime();	// (1970, 1, 1, 1, 1, 1) = 32461 possible bug east of gmt
	m_lastView = GetDefaultTime();	// (2000, 1, 1, 1, 1, 1) = DEF_TIME 946717261
	m_totalViewTime = 0;			// seconds
	m_inMultiTopics = FALSE;
}

void CHistoryItem::Read(CArchive & ar)
{
	m_video.Read(ar);
	ar >> m_lastView;
	ar >> m_totalViewTime;
	ar >> m_inMultiTopics;
}

void CHistoryItem::Write(CArchive & ar)
{
	m_video.Write(ar);
	ar << m_lastView;
	ar << m_totalViewTime;
	ar << m_inMultiTopics;
}



//////////////////////////////////////////////////////////////
// CHistory

CHistory::CHistory()
{
	// Allocate array.
	m_histAr.SetSize(0, 32);

	// Set current schema.
	//
	// v0.8.0.0
	m_schema = 1;
}

CHistory::~CHistory()
{
}

CHistory::CHistory(const CHistory & hist)
{
	Copy(hist);
}
void CHistory::operator=(const CHistory & hist)
{
	Copy(hist);
}
BOOL CHistory::operator==(const CHistory & hist)
{
	// Check schema.
	if(m_schema != hist.m_schema)
		return FALSE;

	// Check array sizes.
	int sz = m_histAr.GetSize();
	if(hist.m_histAr.GetSize() != sz)
		return FALSE;

	// Check each history item.
	for(int i = 0; i < sz; i ++)
	{
		if(m_histAr[i] != hist.m_histAr[i])
			return FALSE;
	}
	// Ok
	return TRUE;
}

BOOL CHistory::operator!=(const CHistory & hist)
{
	return !(*this == hist);
}

void CHistory::Copy(const CHistory & hist)
{
	m_histAr.Copy(hist.m_histAr);
	m_schema = hist.m_schema;
}


// Add or update a History Item.
// Hist param is for a single Video viewing session.
void CHistory::AddHistoryItem(CHistoryItem & hist)
{
	int index;

	// If history item exists, update. 
	if(HistoryExists(hist, index))
	{
		// Update.
		CHistoryItem & item = m_histAr[index];

		// Accumulate time.
		item.m_totalViewTime += hist.m_totalViewTime;

		// Last viewed is this one,
		item.m_lastView = hist.m_lastView;

		// If this video is listed under other topics,
		// update their totalViewTime to match.
		if(item.m_inMultiTopics)
		{
			// Must do a complete linear search.
			int sz = m_histAr.GetSize();
			for(int i = 0; i < sz; i++)
			{
				// Skip history param.
				if(i == index)
					continue;

				// If a match.
				if(m_histAr[i].m_video.title == item.m_video.title)
					m_histAr[i].m_totalViewTime = item.m_totalViewTime;
			}
		}
	}

	// Else add new history item.
	else
	{
		// If viewed for < five seonds, don't add to history.
		if(hist.m_totalViewTime < 5)
			return;

		// Check for duplicate videos in other topics.
		BOOL duplicates = FALSE;
		CDoc * pDoc = ((CApp*)AfxGetApp())->m_pDoc;
		if(pDoc)
		{
			// Find all Topics which also list this video.
			// Store found topics in the topics array.
			CTopicTree & tree = pDoc->GetTree();
			CStringArray topics;
			if(tree.VideoInstances(hist.m_video.title, topics))
			{
				// Temp history item.
				CHistoryItem newHist;

				// If in other topics.
				int sz = topics.GetSize();
				for(int i = 0; i < sz; i++)
				{
					// Skip original.
					if(topics[i] == hist.m_video.topic)
						continue;

					// Build duplicate.
					newHist = hist;

					// Modify last viewed to some 
					// earlier time.
					CTimeSpan t(0, 0, i + 1, 0);	// index plus one minutes
					newHist.m_lastView -= t;		// last viewed n min earlier.

					// Modify Topic to duplicate's topic.
					newHist.m_video.topic = topics[i];

					// Set member flag.
					newHist.m_inMultiTopics = TRUE;

					// Add to array.
					m_histAr.Add(newHist);

					// Set temp flag used below.
					duplicates = TRUE;
				}
			}
		}

		// Add original history item.
		if(duplicates)
			hist.m_inMultiTopics = TRUE;
		m_histAr.Add(hist);

		// Sort - see HistoryExists().
		m_histAr.Sort();
	}
}

// Add a History Item user claims to have viewed.
void CHistory::AddHistoryItemCheater(const CVideo & video)
{
	CHistoryItem hist;
	hist.m_video = video;

	// Set the view time to 100% (seconds).
	int vidTime = _ttoi(hist.m_video.duration);
	hist.m_totalViewTime = vidTime;

	// Set date watched to now.
	hist.m_lastView = CTime::GetCurrentTime();

	// Add.
	AddHistoryItem(hist);
}

// Delete a History Item.
void CHistory::DeleteHistory(const CVideo & video) 
{
	// Get the history item.
	CHistoryItem hist;
	int index;
	if(GetHistoryItem(video, hist, index))
	{
		// If video listed in multiple topics.
		if(hist.m_inMultiTopics)
		{
			// Must do a complete linear search.
			// Note we delete from bottom up.
			int sz = m_histAr.GetSize();
			for(int i = sz - 1; i >= 0; i--)
			{

				// If a match.
				if(m_histAr[i].m_video.title == video.title)
					m_histAr.RemoveAt(i, 1);
			}
		}
		// Else just one history to delete.
		else
			m_histAr.RemoveAt(index, 1);
	}
}

// Get a CHistory Item's m_totalViewTime and m_lastView member values
// formatted for display.
BOOL CHistory::GetHistoryDisplayData(const CVideo & video, 
									 double & timesViewed, CTime & lastViewed)
{
	CHistoryItem hist;
	int index = 0;
	hist.m_video = video;

	if(HistoryExists(hist, index))
	{
		// Get a ref.
		CHistoryItem & item = m_histAr[index];

		// Calc times viewed as a percentage of total viewing time.
		double duration = (double)_ttoi(item.m_video.duration);
		timesViewed = (double)item.m_totalViewTime; 
		timesViewed /= duration;
		// round up
		if(floor(timesViewed + .1) > floor(timesViewed) )
			timesViewed = floor(timesViewed + .1);

		// Copy last viewed CTime.
		lastViewed = item.m_lastView;

		return TRUE;
	}
	return FALSE;
}

// Get an existing CHistory Item for a video, along with its array index.
// Return false if no history for the video.
BOOL CHistory::GetHistoryItem(const CVideo & video, CHistoryItem & hist, int & itemIndex)
{
	CHistoryItem temp;
	int index = 0;
	temp.m_video = video;

	if(HistoryExists(temp, index))
	{
		// Get a ref.
		hist = m_histAr[index];
		itemIndex = index;
		return TRUE;
	}
	return FALSE;
}

BOOL CHistory::BeenViewed(const CVideo & video)
{
	CHistoryItem hist;
	int index = 0;
	hist.m_video = video;

	return HistoryExists(hist, index);
}

// Check if a history exists.
// If existing, init array index of history item.
//
// Note Video Topics and Titles must be unique
// else the first duplicate gets returned.
BOOL CHistory::HistoryExists(CHistoryItem & hist, int & index)
{
	int sz = m_histAr.GetSize();
	index = 0;

	// Loop the m_histAr till we match topics.
	// Note array is sorted by Topic.
	for(int i = 0; i < sz; i++)
	{
		// If at the start of matching topic section.
		if(m_histAr[i].m_video.topic == hist.m_video.topic)
		{
			// Check for Title match within Topic section.
			while(i < sz &&
				  m_histAr[i].m_video.topic == hist.m_video.topic)
			{
				if(m_histAr[i].m_video.title == hist.m_video.title)
				{
					index = i;
					return TRUE;
				}

				// Get next title in this Topic chunk.
				i ++;
			}
			// No match under this Topic.
			return FALSE;
		}
		// Check next for Topic match.
		else
			continue;
	}
	// no go.
	return FALSE;
}

// Same as above but with CVideo param.
// 10/29/13 also check mp4
BOOL CHistory::HistoryExists(CVideo & video, int & index)
{
	int sz = m_histAr.GetSize();
	index = 0;

	// Loop the m_histAr till we match topics.
	// Note array is sorted by Topic.
	for(int i = 0; i < sz; i++)
	{
		// If at the start of matching topic section.
		if(m_histAr[i].m_video.topic == video.topic)
		{
			// Check for Title match within Topic section.
			while(i < sz &&
				  m_histAr[i].m_video.topic == video.topic)
			{
				if(m_histAr[i].m_video.title == video.title &&
				   m_histAr[i].m_video.CompareMp4Strings(video.mp4) == 0 )
				{
					index = i;
					return TRUE;
				}

				// Get next title in this Topic chunk.
				i ++;
			}
			// No match under this Topic.
			return FALSE;
		}
		// Check next for Topic match.
		else
			continue;
	}
	// no go.
	return FALSE;
}

// Public version of above - no index returned.
BOOL CHistory::HistoryExists(CVideo & video)
{
	int index;
	return HistoryExists(video, index);
}

// Check if any history with vidTitle exists.
// Requires linear search.
BOOL CHistory::VideoTitleExists(const CString & vidTitle)
{
	int sz = m_histAr.GetSize();
	for(int i = 0; i < sz; i++)
	{
		// If titles match.
		if(m_histAr[i].m_video.title == vidTitle)
			return TRUE;
	}
	return FALSE;
}

// Check if a History's mp4 in this array matches the vid
// params mp4.
BOOL CHistory::VideoMp4Exists(const CVideo & vid)
{
	int sz = m_histAr.GetSize();
	for(int i = 0; i < sz; i++)
	{
		// dbg
		//CVideo & v = m_histAr[i].m_video;

		// test mp4 only
		// If mp4 match.
		if(m_histAr[i].m_video.CompareMp4Strings(vid.mp4) == 0)
			return TRUE;

		// If titles match.
	//	if(m_histAr[i].m_video.title == vid.title ||
	//	   m_histAr[i].m_video.mp4 == vid.mp4	  )
	//		return TRUE;
	}
	return FALSE;
}


// Count videos viewed in a Topic.
int CHistory::TopicVideosViewed(const CString & treeTopic)
{
	int sz = m_histAr.GetSize();
	int nVidsViewed = 0;

	// Loop the m_histAr till we match topics.
	// Note array is sorted by Topic.
	for(int i = 0; i < sz; i++)
	{
		// If at the start of matching topic section.
		if(m_histAr[i].m_video.topic == treeTopic)
		{
			// Check for Title match within Topic section.
			while(i < sz && m_histAr[i].m_video.topic == treeTopic)
			{
				// If any view time.
				if(m_histAr[i].m_totalViewTime)
					nVidsViewed ++;

				// Get next title in this Topic chunk.
				i++;
			}
			// No match uner this Topic.
			return nVidsViewed;
		}
	}
	// None.
	return 0;
}


// Set lastViewed param with the last viewed Video.
// Return false if no history.
// Performs a linear search comparing single CTime int.
BOOL CHistory::GetLastViewedVideo(CVideo & lastViewed)
{
	int sz = m_histAr.GetSize();
	if(!sz)
		return FALSE;

	CTime time(DEF_TIME);//(2000, 1, 1, 1, 1, 1); // any time pre app.
	int arElmnt = -1;

	//dbg
	//CTime t = time;

	// Loop all.
	for(int i = 0; i < sz; i++)
	{
		// Get earliest.
		if(m_histAr[i].m_lastView > time)
		{
			time = m_histAr[i].m_lastView;
			arElmnt = i;

			// dbg
			//t = time;
		}
	}

	// ArElmnt is the one.
	if(arElmnt >= 0 && arElmnt < sz)
	{
		lastViewed = m_histAr[arElmnt].m_video;
		return TRUE;
	}
	// Problem.
	return FALSE;
}

// Save history to a backup.
BOOL CHistory::BackupHistory(CString bkPath)
{
	// Write to backup file.
	return Write(bkPath);
}

// Note bug: Histories created in vs2005 will not restore
// in VC6. Must be some alignment issue betwwen the 
// file i/o as CArchive is choaking. Would be nice to 
// figured this out.
// May be with CTime #define (_USE_32BIT_TIME_T) - vs2005 is 64 bit CTime.
// Most likely, see:
// http://msdn.microsoft.com/en-us/library/w4ddyt9h%28v=vs.80%29.aspx
// Note however the vs2005 version accepts archives from VC6,
// so I'm gonna forget about it.
//
// Restore history from a backup.
BOOL CHistory::RestoreHistory(CString bkPath)
{
	// Temp CHistory.
	CHistory temp;
	CString retMsg, msg;

	// Read from bkPath into a temp CHistory
	// so we can validate.
	if(temp.Read(bkPath))
	{
		// Ensure something read.
		if(temp.NumHistoryItems() == 0)
		{
			retMsg = _T("Backup file is empty.\n");
			MessageBox(NULL, retMsg, _T("History Restore"), MB_OK | MB_ICONINFORMATION);
			return FALSE;
		}

		// Validate the temp history.
		int ret = temp.Validate(retMsg);

		// Error
		if(ret == 0)
		{
			msg.Format(_T("Restoration failed\n%s"), retMsg);
			MessageBox(NULL, msg, _T("History Restore"), MB_OK | MB_ICONINFORMATION);
			return FALSE;
		}
		// Warning
		else if(ret == 2)
		{
			msg.Format(_T("Restored History required repair:\n%sAccept changes?"), retMsg);
			int ret = MessageBox(NULL, msg, _T("History Restore"), 
								 MB_YESNO | MB_ICONINFORMATION);
			// Bail per user choice.
			if(ret == IDNO)
				return FALSE;
		}
		// All good.
		else if(ret == 1)
			MessageBox(NULL, retMsg, _T("History Restore"), MB_OK | MB_ICONINFORMATION);

		// Copy restored and validated temp CHistory
		// array to program array.
		//
		// Clean existing.
		m_histAr.RemoveAll();

		// Copy.
		m_histAr.Copy(temp.m_histAr);

		// Sort
		m_histAr.Sort();

		// Done
		return TRUE;
	}
	// Error file i/o
	else
	{
		retMsg = _T("History restore failed.\n");
		MessageBox(NULL, retMsg, _T("History Restore"), MB_OK | MB_ICONINFORMATION);
	}
	return FALSE;
}

// consider
BOOL CHistory::MergeHistory(CString bkPath)
{
	// Temp CHistory.
	CHistory toMerge, newHist;
	CString retMsg, msg;

	// Read from bkPath into a temp CHistory.
	if(toMerge.Read(bkPath))
	{
		// Ensure something read.
		if(toMerge.NumHistoryItems() == 0)
		{
			retMsg = _T("Backup history file for merging is empty.\n");
			MessageBox(NULL, retMsg, _T("Merge History"), MB_OK | MB_ICONINFORMATION);
			return FALSE;
		}

		// Merge with current history.
		int ret = Merge(toMerge, newHist, retMsg);

		// Error or merge hist needs repair
		if(ret == 0)
		{
			msg.Format(_T("Merge failed\n%s"), retMsg);
			MessageBox(NULL, msg, _T("Merge History"), MB_OK | MB_ICONINFORMATION);
			return FALSE;
		}
		// Nothing to merge.
		else if(ret == 1)
		{
	//		msg.Format(_T("Nothing to Merge\n%s"), retMsg);
			MessageBox(NULL, retMsg, _T("Merge History"), MB_OK | MB_ICONINFORMATION);
			return FALSE;
		}
		// Items merged.
		else if(ret == 2)
		{
			msg.Format(_T("Merge History successfull:\n%s\nAccept changes?"), retMsg);
			int ret = MessageBox(NULL, msg, _T("Merge History"), 
								 MB_YESNO | MB_ICONINFORMATION);
			// Bail per user choice.
			if(ret == IDNO)
				return FALSE;
		}

		// Copy restored and validated temp CHistory
		// array to program array.
		//
		// Clean existing.
		m_histAr.RemoveAll();

		// Copy.
		m_histAr.Copy(newHist.m_histAr);

		// Sort
		m_histAr.Sort();

		// Done
		return TRUE;
	}
	// Error file i/o
	else
	{
		retMsg = _T("Merge History failed.\n");
		MessageBox(NULL, retMsg, _T("Merge History"), MB_OK | MB_ICONINFORMATION);
	}
	return FALSE;
}


// Repair this CHistory.
BOOL CHistory::RepairHistory()
{
	// Temp CHistory.
	CString retMsg, msg;

	// Ensure somethin to repair..
	if(NumHistoryItems() == 0)
	{
		retMsg = _T("History is empty.\n");
		MessageBox(NULL, retMsg, _T("History Repair"), MB_OK | MB_ICONINFORMATION);
		return FALSE;
	}

	// Copy this CHistory to a temp CHistory.
	CHistory temp;
	temp.m_histAr.Copy(m_histAr);

	// Ensure something copied.
	if(temp.NumHistoryItems() == 0)
	{
		retMsg = _T("Error repairing History.\n");
		MessageBox(NULL, retMsg, _T("History Repair"), MB_OK | MB_ICONINFORMATION);
		return FALSE;
	}

	// Validate the temp history.
	int ret = temp.Validate(retMsg);

	// Error
	if(ret == 0)
	{
		msg.Format(_T("History Repair failed\n%s"), retMsg);
		MessageBox(NULL, msg, _T("History Repair"), MB_OK | MB_ICONINFORMATION);
		return FALSE;
	}
	// Warning
	else if(ret == 2)
	{
		msg.Format(_T("History required repair:\n%sAccept changes?"), retMsg);
		int ret = MessageBox(NULL, msg, _T("History Repair"), 
							 MB_YESNO | MB_ICONINFORMATION);
		// Bail per user choice.
		if(ret == IDNO)
			return FALSE;
	}
	// All good.
	else if(ret == 1)
		MessageBox(NULL, retMsg, _T("History Repair"), MB_OK | MB_ICONINFORMATION);

	// Copy restored and validated temp CHistory
	// array to program array.
	//
	// Clean existing.
	m_histAr.RemoveAll();

	// Copy.
	m_histAr.Copy(temp.m_histAr);

	// Sort
	m_histAr.Sort();

	// Done
	return TRUE;
}

// Verify this CHistory is valid for the current KhanDB Topic Tree.
//
// Used when the topicTree has been updated or when a CHistory 
// is being restored.
//
// Typical usage would be to copy the CHistory in question,
// then have the copy call this function. The validated
// copy can then be utilized.
//
// Return:
// 0 for failure
// 1 for no changes
// 2 for repaired 
int CHistory::Validate(CString & msg)
{
	CVideoArray videoAr;
	CVideoArray videosToAdd;
	videosToAdd.SetSize(0, 32);
	CHistory tempHist;
	CHistoryItem tempHistItem;
	CString vidTitle, s;
	int index = 0;
	int nInvalidVideos = 0;
	int nRepairedHistories = 0;
	int nUnchanged = 0;
	BOOL repaired = FALSE;
	BOOL unchanged = FALSE;
	int ret = 0;

	// Get the current topic tree.
	CDoc * pDoc = ((CApp*)AfxGetApp())->m_pDoc;
	if(!pDoc)
		return 0;
	CTopicTree & tree = pDoc->GetTree();

	// Count unique history items.
	int unique = NumUniqueHistoryItems();

	// Add each valid CHistoryItem to the tempHist array.
	int histSz = m_histAr.GetSize();
	for(int iHist = 0; iHist < histSz; iHist ++)
	{
		repaired = FALSE;
		unchanged = FALSE;

		//dbg
		if(iHist == 6)
			int g = 0;

		// Skip duplicates. Note code below adds all possible
		// histories associated with an mp4.
		if(tempHist.VideoMp4Exists(m_histAr[iHist].m_video))
			continue;

		// Get all the videos in the current topicTree matching
		// this history item's video mp4. Store in the videoAr 
		// param.
		//
		// Note these videos will contain the current valid 
		// Topic Path, Title, and Mp4.
		if(tree.VideoInstancesMp4(m_histAr[iHist].m_video, videoAr))
		{
			// Check the valid videos against this history.
			int vidSz = videoAr.GetSize();
			for(int iVid = 0; iVid < vidSz; iVid ++)
			{
				// Reuse any history with valid video Topic path, 
				// Title, and Mp4.
				if(HistoryExists(videoAr[iVid], index))
				{
					// dbg
					//CVideo vid1 = videoAr[iVid];
					//CVideo vid2 = m_histAr[index].m_video;

					tempHist.m_histAr.Add(m_histAr[index]);
					unchanged = TRUE;
				}

				// Else create and add a new history based on the 
				// existing invalid one, using the new Video's
				// valid Topic path, Title and Mp4.
				else
				{
					tempHistItem = m_histAr[iHist];
					tempHistItem.m_video = videoAr[iVid];
					tempHistItem.m_inMultiTopics = vidSz > 1 ? TRUE : FALSE;

					// Modify lastViewed time for multiples.
					if(iVid)
					{
						// Get a one minute time span.
						CTimeSpan t(0, 0, iVid + 1, 0);
						// Set last viewed one minute earlier.
						tempHistItem.m_lastView -= t;
					}
					// Add.
					tempHist.m_histAr.Add(tempHistItem);

					// This History item, along with it's multiples,
					// has been repaired
					repaired = TRUE;
				}
			}
		}
		else
			nInvalidVideos ++;

		// Count repaired unique videos.
		if(repaired)
			nRepairedHistories ++;

		// Count unchanged - not unique, includes existing history multiples.
		if(unchanged)
			nUnchanged ++;
	}
	// Get n copied.
	int newSz = tempHist.NumHistoryItems();

	// If none.
	if(!newSz)
	{
		msg = _T("No valid History items were found.");
		return 0;
	}

	// Copy valid items to array.
	m_histAr.Copy(tempHist.m_histAr);

	// Sort - see HistoryExists().
	m_histAr.Sort();

	// Count unique history items.
	int unique2 = NumUniqueHistoryItems();

	// Set msgs and return value.
	ret = 1;
	if(nRepairedHistories)
	{
		s.Format(_T("%d of %d History items were repaired.\n"), 
						nRepairedHistories, unique);
		msg += s;
		ret = 2;
	}
	if(nInvalidVideos)
	{
		s.Format(_T("%d of %d History items were discarded - videos not found.\n"), 
						nInvalidVideos, unique);
		msg += s;
		ret = 2;
	}
	if(ret == 2)
	{
		s.Format(_T("History contains %d valid items.\n"), 
						unique2);
		msg += s;
	}

	if(ret == 1)
	{
		s.Format(_T("All History items are valid.\nHistory contains %d items.\n"), unique2);
		msg += s;
	}
	return ret;
}

// Merge mergeHistory param with this history.
// Save merged history in newHistory param.
//  Return:
//  0 for failure
//  1 for no no merges needed
//  2 items merged
int CHistory::Merge(CHistory & mergeHistory, CHistory & newHistory, CString & msg)
{
	CVideoArray videoAr;
	CVideoArray videosToAdd;
	videosToAdd.SetSize(0, 32);
	CHistoryItem tempHistItem;
	CString vidTitle, s;
	int index = 0;
	int nNotMergeable = 0;
	int nMergedHistories = 0;
	int nUnchanged = 0;
	BOOL merged = FALSE;
	BOOL unchanged = FALSE;
	int iHist, ret = 0;

	// Copy current history to newHist param.
	newHistory.Copy(*this); 

	// Ensure the current history is not in need of repair.
	// Note items in the mergeHistory param are "repaired" 
	// below as needed.
	ret = newHistory.Validate(msg);
	if(ret != 1)
	{
		msg = _T("Your current history needs to be repaired first.\n")
			  _T("Please use the Hisory/Repair menu item to repair.");
		return 0;
	}

	// Get the current topic tree.
	CDoc * pDoc = ((CApp*)AfxGetApp())->m_pDoc;
	if(!pDoc)
		return 0;
	CTopicTree & tree = pDoc->GetTree();

	// Count unique history items in current history.
	int nExistUnique = NumUniqueHistoryItems();

	// Get a reference to the merge array.
	CHistoryArray & mergeHistAr = mergeHistory.m_histAr;

	// Add each valid merge CHistoryItem to the tempHist array.
	int histSz = mergeHistAr.GetSize();
	for(iHist = 0; iHist < histSz; iHist ++)
	{
		merged = FALSE;

		// Skip if merg item already exists in the current history.
		if(newHistory.VideoMp4Exists(mergeHistAr[iHist].m_video))
			continue;

		// Else get all videos in the current db with the merge item's mp4.
		if(tree.VideoInstancesMp4(mergeHistAr[iHist].m_video, videoAr))
		{
			// Add a new history item for valid video based on
			// the merge history item.
			int vidSz = videoAr.GetSize();
			for(int iVid = 0; iVid < vidSz; iVid ++)
			{
				tempHistItem = mergeHistAr[iHist];
				tempHistItem.m_video = videoAr[iVid];
				tempHistItem.m_inMultiTopics = vidSz > 1 ? TRUE : FALSE;

				// Modify lastViewed time for multiples.
				if(iVid)
				{
					// Get a one minute time span.
					CTimeSpan t(0, 0, iVid + 1, 0);
					// Set last viewed one minute earlier.
					tempHistItem.m_lastView -= t;
				}
				// Add.
				newHistory.m_histAr.Add(tempHistItem);

				// This History item, along with it's multiples,
				// has been merged.
				merged = TRUE;
			}
		}
		else
			nNotMergeable ++;

		// Count merged videos.
		if(merged)
			nMergedHistories ++;
	}

	// If merged history items.
	int nMerge = newHistory.NumUniqueHistoryItems();
	if(nMergedHistories && nExistUnique + nMergedHistories == nMerge)
	{
		msg.Format(_T("%d history items merged."), nMergedHistories);
		return 2;
	}
	else if(!nMergedHistories)
	{
		msg.Format(_T("No history items needed merging."));
		return 1;
	}
	else
		msg.Format(_T("Could not perform merge."));
		
	return 0;
}

// Count unique history items based on their video's mp4.
int CHistory::NumUniqueHistoryItems()
{
	if(!m_histAr.GetSize())
		return 0;

	// Make a copy for sorting by mp4.
	CHistoryArray temp;
	temp.Copy(m_histAr);

	// Sort by mp4.
	temp.SortMp4Count();

	// Sorted count.
	int sz = temp.GetSize();
	if(!sz)
		return 0;

	// The first item counts as unique.
	int nItems = 1;

	// Check the rest.
	for(int i = 1; i < sz; i++)
	{
		// Unique if non zero.
		if(temp[i-1].m_video.CompareMp4Strings(temp[i].m_video.mp4) != 0)
			nItems ++;
	}
	return nItems;
}

//////////////////////////////////////////////////////////////
// Serialize

// Read from default store.
BOOL CHistory::Read()
{
	int ret = TRUE; 

	// Get user data directory from CApp.
//	CString path = ((CApp*)AfxGetApp())->GetUserSettingsFolder();
	CString path;
	if( ((CApp*)AfxGetApp())->GetUserSettingsFolder(path))
	{
		path += _T("\\");
		path += HIST_STORE_NAME;
	}
	else
		return FALSE;
//	path += _T("\\");
//	path += HIST_STORE_NAME;
	return Read(path);
}

BOOL CHistory::Read(CString path)
{
	int ret = TRUE; 

	// Open the file for reading.
	CFile file;
	CFileException ex;
	if(!file.Open(path, CFile::modeRead | CFile::shareDenyWrite, &ex))
	{
	//	ex.ReportError();
		return FALSE;
	}
	if(file.GetLength() == 0)
	{
	//	msg.Format(_T("%s is an empty file"), filePathShort);
	//	AfxMessageBox(msg);
		TRACE(_T("ERROR CHistory Read GetLength() = 0\n"));
		return FALSE;
	}

	CArchive ar(&file, CArchive::load);
	TRY
	{	
		// Get our schema
		int schema; // note GetObjectSchema() is for serializable classes
		ar >> schema;
		if(schema == 1)
		{
			// Read
			int sz;
			ar >> sz;
			CHistoryItem hist;
			for(int i = 0; i < sz; i++)
			{
				hist.Read(ar);
				m_histAr.Add(hist);
				hist.Zero();
			}

			// Close the archive.	
			ar.Close();
			// Close the file	
			file.Close();
		}
		// Else an error
		else
		{
			AfxMessageBox(_T("KDHist.dat is an invalid Khan Video History version."));
			
			// Close Archive
			ar.Close();
			// Close the file and ignore any assertions	
			file.Abort();
			return FALSE;
		}
	}
	CATCH_ALL(e)
	{
		// Close the archive.	
		ar.Close();
		// Close file safely and quietly
		file.Abort(); 
		TRY
		{
			// CDoc member
			//ReportSaveLoadException(filePath, e, FALSE, AFX_IDP_FAILED_TO_OPEN_DOC);
		}
		END_TRY
		e->Delete();
		ret = FALSE;
	}
	END_CATCH_ALL

	return ret;
}


// Write to default store.
BOOL CHistory::Write()
{
	// Get user data directory from CApp.
//	CString path = ((CApp*)AfxGetApp())->GetUserSettingsFolder();
	CString path;
	if( ((CApp*)AfxGetApp())->GetUserSettingsFolder(path))
	{
		path += _T("\\");
		path += HIST_STORE_NAME;
	}
	else
		return FALSE;

//	path += _T("\\");
//	path += HIST_STORE_NAME;
	return Write(path);
}

BOOL CHistory::Write(CString path)
{
	int ret = TRUE;

	// Open the file for writing.
	CFile file;
	CFileException ex;
	if(!file.Open(path, CFile::modeCreate | CFile::modeWrite , &ex))
	{
		ex.ReportError();
		return FALSE;
	}

	CArchive ar(&file, CArchive::store);
	TRY
	{
		// Write per our schema.
		if(m_schema == 1)
		{
			// Set schema.
			ar << m_schema; // note SetObjectSchema() is for serializable classes

			// Write.
			int sz = m_histAr.GetSize();
			ar << sz;
			for(int i = 0; i < sz; i++)
			{
				m_histAr[i].Write(ar);
			}

		}
		// Next Schema.
		// .....

		// Clean up.
		//
		// Close the archive.	
		ar.Close();
		// Close the file	
		file.Close();
	}
	CATCH_ALL(e)
	{
		// Close the archive.	
		ar.Close();
		// Close file safely and quietly
		file.Abort(); 
		TRY
		{
			// CDoc member
			//ReportSaveLoadException(curDir, e, TRUE, AFX_IDP_FAILED_TO_SAVE_DOC);
			// THROW_LAST();
		}
		END_TRY
		e->Delete();
		ret = FALSE;
	}
	END_CATCH_ALL

	return ret;
}


//////////////////////////////////////////////////////////////
// CHistoryArray - soratable CArray

// Sort the History.
void CHistoryArray::Sort() 
{
	qsort(GetData(), GetSize(), sizeof(CHistoryItem), SortHistory);
}

int CHistoryArray::SortHistory(const void *arg1, const void *arg2) 
{
	CHistoryItem hist1 = *(CHistoryItem *)arg1;
	CHistoryItem hist2 = *(CHistoryItem *)arg2;

	// Topics first - alphabetically ascending.
	int ret = hist1.m_video.topic.CompareNoCase(hist2.m_video.topic);

	// Last viewed next - time ascending.
	if(ret == 0)
	{
		if(hist1.m_lastView < hist2.m_lastView)
			ret = -1;
		else if(hist1.m_lastView > hist2.m_lastView)
			ret = +1;
	}
	return ret;
}

// Sort the History by mp4.
// Used only on copies of the History
// to establish unique history items.
void CHistoryArray::SortMp4Count() 
{
	qsort(GetData(), GetSize(), sizeof(CHistoryItem), SortHistoryMp4Count);
}

int CHistoryArray::SortHistoryMp4Count(const void *arg1, const void *arg2) 
{
	CHistoryItem hist1 = *(CHistoryItem *)arg1;
	CHistoryItem hist2 = *(CHistoryItem *)arg2;

	// Sort by mp4 alphabetically ascending.
//	return hist1.m_video.mp4.CompareNoCase(hist2.m_video.mp4);
	return hist1.m_video.CompareMp4Strings(hist2.m_video.mp4);
}

